/* Share X Window System selections with a console using gpm.

   Copyright 2001, 2002, 2007, 2008, 2010, 2013, 2015, 2020 Kevin Ryde.

   This file is part of x2gpm.

   x2gpm is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   x2gpm is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with X2gpm; see the file COPYING.  Failing that, goto
   <http://www.gnu.org/licenses/>. */


#define _GNU_SOURCE
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <locale.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/types.h>
#include <sys/poll.h>

#undef WUNTRACED  /* avoid redefines from <linux/wait.h> */
#undef WNOHANG
#undef WSTOPPED
#undef WEXITED
#undef WCONTINUED
#include <linux/keyboard.h>  /* for KG_CTRL */
#include <linux/kd.h>        /* for KDGKBMODE */

#include <gpm.h>

#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


#define VERSION 11


int   option_all = 0;
char  *option_display = NULL;
char  *option_pidfile = NULL;
int   option_sync = 0;
int   option_verbose = 0;
int   option_debug = 0;
int   option_foreground = 0;

char **saved_argv;
Gpm_Connect conn;

void
beep (void)
{
  write (1, "\a", 1);
}

/* console_filename is "/dev/tty" for single screen, or "/dev/console" for
   "-a" all mode.  console_fd is the opened console_filename.
   In single mode, console_fd is kept open permanently.
   In "-a" all mode, it's reopened for each operation to get the current
   screen.
   ensure_current_console() opens console_fd if not already open.
   finished_current_console() closes it if "-"a mode and if not already
   closed, or does nothing for single mode.
*/
int console_fd = -1;
const char *console_filename;
void
ensure_current_console (void)
{
  if (console_fd == -1)
    {
      console_fd = open (console_filename, O_WRONLY);
      if (console_fd == -1)
        {
          syslog (LOG_ERR, "cannot open %s: %s",
                  console_filename, strerror(errno));
          beep ();
          exit (1);
        }
    }
}
void
finished_current_console (void)
{
  if (option_all && console_fd != -1)
    {
      if (close (console_fd) != 0)
        syslog (LOG_ERR, "error closing %s: %s",
                console_filename, strerror(errno));
      console_fd = -1;
    }
}

/* return true if PRIORITY is wanted by the current setlogmask() */
int
logpriority_wanted (int priority)
{
  int mask = setlogmask (0);
  setlogmask (mask);
  return (mask & LOG_MASK(priority));
}

/* gcc 2.5 and up for attribute printf, according to glib at least */
#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 4)
void x2gpm_log (int priority, const char *fmt, ...)
  __attribute__ ((format (printf, 2, 3)));
#endif

void
x2gpm_log (int priority, const char *fmt, ...)
{
  {
    va_list ap;
    va_start (ap, fmt);
    vsyslog (priority, fmt, ap);
    va_end (ap);
  }

  if ((priority == LOG_ERR || option_verbose)
      && logpriority_wanted (priority))
    {
      static const char pre[] = "[x2gpm: ";
      static const char post[] = "]\n";
      va_list ap;

      ensure_current_console ();
      write (console_fd, pre, sizeof(pre)-1);
      va_start (ap, fmt);
      vdprintf (console_fd, fmt, ap);
      va_end (ap);
      write (console_fd, post, sizeof(post)-1);
    }
}

int
console_is_utf8 (void)
{
  static int warned = 0;
  long mode;

  ensure_current_console ();
  if (ioctl (console_fd, KDGKBMODE, &mode) != 0)
    {
      /* error will be something fishy, maybe an ancient kernel? */
      if (! warned)
        {
          x2gpm_log (LOG_ERR,
                     "ioctl KDGKBMODE %s error, assuming not utf8: %s",
                     console_filename, strerror(errno));
          warned = 1;
        }
      return 0;
    }
  return (mode == K_UNICODE);
}

char *
x11_text_conv_strerror (int ret)
{
  switch (ret) {
  case XNoMemory:            return "no memory";
  case XLocaleNotSupported:  return "locale not supported";
  case XConverterNotFound:   return "converter not found";
  default:                   return "unknown error";
  }
}

/* There's no apparent way to make Xlib recover gracefully from a server
   shutdown.  We can know that EOF has been reached on the server
   descriptor, but a call to XCloseDisplay tries to do some sort of XSync
   round trip which bombs.  XSetIOErrorHandler doesn't help, it can't return
   or anything.  */
int
x_io_error_handler (Display *display)
{
  execv (saved_argv[0], saved_argv);
  x2gpm_log (LOG_ERR, "oops, cannot re-exec for X error recovery");
  exit (1);
}
void
restart (void)
{
  x_io_error_handler (NULL);
}

char *xauthority_info = "";
char *xauthority_env = NULL;

/* Put an XAUTHORITY environment variable in the environment, according to
   the user logged in at option_display.  Return 1 if successful, or 0 if
   nobody logged in.  */
void
setup_xauthority (void)
{
  struct utmp want, *got;
  struct passwd *pw;
  char *env;

  /* just .Xauthority file when in single mode */
  if (! option_all)
    return;

  /* glibc takes the given pointer into the environment.  After unsetting
     it's safe to free that previous pointer, held in xauthority_env.  */
  unsetenv ("XAUTHORITY");
  if (xauthority_env)
    {
      free (xauthority_env);
      xauthority_env = NULL;
    }

  /* In the past utmp had ut_line ":0" display name.  But circa May 2015 it
     has ut_line "console" and ut_id ":0".  Suspect dodginess in sessreg.
     Look for either way.
  */
  setutent();
  memset (&want, '\0', sizeof(want));
  strncpy (want.ut_line, option_display, sizeof(want.ut_line)-1);
  got = getutline (&want);
  endutent();
  if (got == NULL)
    {
      setutent();
      memset (&want, '\0', sizeof(want));
      strncpy (want.ut_line, "console", sizeof(want.ut_line)-1);
      strncpy (want.ut_id, option_display, sizeof(want.ut_id)-1);
      got = getutline (&want);
      endutent();

      if (got == NULL)
        {
          xauthority_info = "nobody logged in";
          return;
        }
    }

  pw = getpwnam (got->ut_user);
  if (pw == NULL)
    {
      xauthority_info = "unknown user logged in";
      return;
    }

  if (asprintf (&env, "XAUTHORITY=%s/.Xauthority", pw->pw_dir) < 0)
    {
      x2gpm_log (LOG_ERR, "asprintf error: %s", strerror(errno));
      beep ();
      exit (1);
    }
  xauthority_env = env;

  if (putenv (env) != 0)
    {
      x2gpm_log (LOG_ERR, "putenv error: %s", strerror(errno));
      beep ();
      exit (1);
    }
  xauthority_info = "";
}

/* paste null-terminated "str" into the console using ioctl(TIOCSTI)  */
int
tiocsti (const char *str)
{
  x2gpm_log (LOG_DEBUG, "TIOCSTI: \"%s\"", str);
  ensure_current_console ();

  for ( ; *str != '\0'; str++)
    {
      if (ioctl (console_fd, TIOCSTI, &str[0]) != 0)
        {
          x2gpm_log (LOG_ERR, "%s TIOCSTI error: %s",
                     console_filename, strerror(errno));
          beep ();
          return 0;
        }
    }
  return 1;
}

/* XOpenDisplay() prints some messages to stderr, in particular when the
   server refuses a connection XOpenDisplay() prints the server's reason
   message to stderr.  For example the slightly unenlightening "No protocol
   specified" from the server normally means auth info from XAUTHORITY,
   ~/.Xauthority etc either refused or not supplied.

   Original Xlib prints with fprintf(stderr) from its XOpenDis.c.  xcb
   prints with write(2) from its xcb_conn.c read_setup().  Capture either by
   duping stderr descriptor to a temp file and reading back anything written
   there.
*/
Display *
x_open_display (void)
{
  FILE *temp_file;
  int temp_fd, orig_stderr;
  Display *display;
  ssize_t len;
  char buf[256];

  temp_file = tmpfile (); /* in mode "wb+" */
  if (! temp_file) {
    x2gpm_log (LOG_ERR, "cannot create temp file: %s", strerror(errno));
    return NULL;
  }
  temp_fd = fileno(temp_file);

  orig_stderr = dup (2);
  if (orig_stderr < 0) {
    x2gpm_log (LOG_ERR, "cannot dup stderr: %s", strerror(errno));
    fclose(temp_file);
    return NULL;
  }
  if (dup2(temp_fd,2) < 0) {
    x2gpm_log (LOG_ERR, "cannot dup temp to stderr: %s", strerror(errno));
    fclose(temp_file);
    return NULL;
  }
  display = XOpenDisplay (option_display);
  if (dup2(orig_stderr,2) < 0) {
    x2gpm_log (LOG_ERR, "cannot dup original back to stderr: %s", strerror(errno));
    abort();
  }

  lseek (temp_fd, SEEK_SET, 0L);
  len = read (temp_fd, buf, sizeof(buf)-1);
  if (len > 0) {
    while (len > 0 && buf[len-1] == '\n') {
      len--;  /* strip trailing newlines */
    }
    buf[len] = '\0';
  } else {
    buf[0] = 0;
  }
  if (fclose(temp_file) < 0) {
    x2gpm_log (LOG_ERR, "error closing temp file: %s", strerror(errno));
  }

  if (display)
    return display;

  x2gpm_log (LOG_WARNING,
             "cannot connect to X server %s, %s, %s",
             option_display, xauthority_info, buf);
  return NULL;
}

void
x2gpm (void)
{
  Display     *display = NULL;
  Window      window = 0;
  int         x_connected = 0;
  int         expecting_paste = 0;
  struct pollfd  pollfd[2];

  pollfd[0].fd = gpm_fd;
  pollfd[0].events = POLLIN;
  pollfd[1].events = POLLIN;
  pollfd[1].revents = 0; /* initial */

  for (;;)
    {
      finished_current_console ();
      x2gpm_log (LOG_DEBUG, "poll ... ");
      poll (pollfd, x_connected ? 2 : 1, -1);
      x2gpm_log (LOG_DEBUG, "got gpm revents=%#X, X revents=%#X",
                 pollfd[0].revents, pollfd[1].revents);

      if (pollfd[0].revents)
        {
          Gpm_Event  e;
          int        ret;

          ret = Gpm_GetEvent (&e);
          if (ret == 0)
            {
              x2gpm_log (LOG_ERR, "gpm eof, will attempt restart in 5 seconds");
              sleep (5);
              restart ();
            }
          if (ret != 1)
            {
              x2gpm_log (LOG_ERR, "Gpm_GetEvent error, restarting");
              restart ();
            }

          switch (GPM_BARE_EVENTS (e.type)) {
          case GPM_DOWN:
            x2gpm_log (LOG_DEBUG, "GPM_DOWN buttons=%d clicks=%d",
                       e.buttons, e.clicks);
            if (e.clicks == 0)
              {
                Window owner_window;

                if (! (e.buttons & GPM_B_RIGHT))
                  break;  /* only act on right button */

                if (! x_connected)
                  {
                    setup_xauthority ();

                    display = x_open_display ();
                    if (display == NULL)
                      {
                        beep ();
                        break;
                      }
                    x2gpm_log (LOG_DEBUG, "X connected");
                    x_connected = 1;
                    pollfd[1].fd = ConnectionNumber (display);

                    XSetIOErrorHandler (x_io_error_handler);
                    if (option_sync)
                      XSynchronize (display, True);

                    window = XCreateSimpleWindow (display,
                                                  DefaultRootWindow (display),
                                                  0, 0, 1, 1, 0, 0, 0);

                    /* WM_CLASS for application identification (would be
                       resource etc, but we're never mapped) */
                    {
                      XClassHint classhint;
                      classhint.res_name = "x2gpm";
                      classhint.res_class = "X2gpm";
                      XSetClassHint(display, window, &classhint);
                    }

                    /* WM_NAME for title bar display.  We're never mapped,
                       but might be displayed by something like xrestop. */
                    XStoreName(display, window, "X2gpm");

                    {
                      char hostname[HOST_NAME_MAX];
                      if (gethostname(hostname, sizeof(hostname)) == 0) {
                        Atom wm_client_machine
                          = XInternAtom(display, "WM_CLIENT_MACHINE", 0);
                        XChangeProperty(display, window,
                                        wm_client_machine,
                                        XA_STRING, 8, PropModeReplace,
                                        (unsigned char *) hostname,
                                        strnlen(hostname,sizeof(hostname)));
                      }
                    }
                    {
                      Atom net_wm_pid = XInternAtom(display, "_NET_WM_PID", 0);
                      long pid = (long) getpid();
                      XChangeProperty(display, window,
                                      net_wm_pid,
                                      XA_CARDINAL, 32, PropModeReplace,
                                      (unsigned char *) &pid, 1);
                    }
                  }

                owner_window = XGetSelectionOwner (display, XA_PRIMARY);
                if (owner_window == None)
                  {
                    x2gpm_log (LOG_INFO, "no current X selection");
                    beep ();
                    break;
                  }

                /* We're going to ask the selection to be written to
                   property XA_PRIMARY on our window.  ICCCM version 2.0
                   says that destination window property should not exist.
                   (A new requirement in ICCCM 2.0).  */
                XDeleteProperty (display, window, XA_PRIMARY);

                /* Target "TEXT" means the owner chooses between STRING or
                   COMPOUND_TEXT.  The Xmb and Xutf8 funcs below handle
                   conversion out of either.  */
                x2gpm_log (LOG_DEBUG, "XConvertSelection from owner win %#lX",
                           (unsigned long) owner_window);
                XConvertSelection (display, XA_PRIMARY,
                                   XInternAtom (display, "TEXT", False),
                                   XA_PRIMARY, window, CurrentTime);
                expecting_paste = 1;
                XFlush (display);
              }
            break;

          default:
            x2gpm_log (LOG_WARNING, "oops, unexpected gpm event %d", e.type);
            break;
          }
        }

      if (x_connected)
        {
          if (pollfd[1].revents & POLLHUP)
            {
              x2gpm_log (LOG_INFO, "X connection lost, restarting");
              restart ();
            }

          if (pollfd[1].revents)
            {
              XEventsQueued (display, QueuedAfterReading);
            }
          x2gpm_log (LOG_DEBUG, "X queue %d events",
                     XEventsQueued (display, QueuedAlready));

          while (XEventsQueued (display, QueuedAlready))
            {
              XEvent        e;
              XTextProperty textprop;
              char          **textlist;
              int           n_textlist;

              XNextEvent (display, &e);

              switch (e.type) {
              case SelectionNotify:
                x2gpm_log (LOG_DEBUG, "SelectionNotify received, expecting_paste=%d, property=%ld",
                           expecting_paste, (long) e.xselection.property);
                if (! expecting_paste)
                  {
                    x2gpm_log (LOG_WARNING,"oops, unexpected SelectionNotify");
                    break;
                  }

                if (e.xselection.property == None && expecting_paste == 1)
                  {
                    /* Target "TEXT" failed, try target "STRING".
                       For example xpdf will give STRING but not TEXT.  */
                    x2gpm_log (LOG_DEBUG, "XConvertSelection TEXT failed, try STRING");
                    XConvertSelection (display, XA_PRIMARY,
                                       XInternAtom (display, "STRING", False),
                                       XA_PRIMARY, window, CurrentTime);
                    XFlush (display);
                    expecting_paste = 2;
                    break;
                  }
                expecting_paste = 0;

                if (e.xselection.property == None)
                  {
                    x2gpm_log (LOG_WARNING, "X selection not converted");
                    beep ();
                    break;
                  }
                if (! XGetTextProperty (display, window,
                                        &textprop, XA_PRIMARY))
                  {
                    x2gpm_log (LOG_WARNING, "XGetTextProperty error");
                    beep ();
                    break;
                  }

                {
                  int ret;
                  const char *conv_func_name;
                  if (console_is_utf8 ())
                    {
                      conv_func_name = "Xutf8TextPropertyToTextList";
                      ret = Xutf8TextPropertyToTextList
                        (display, &textprop, &textlist, &n_textlist);
                    }
                  else
                    {
                      conv_func_name = "XmbTextPropertyToTextList";
                      ret = XmbTextPropertyToTextList
                        (display, &textprop,
                         &textlist, &n_textlist);
                    }

                  if (ret < 0)
                    {
                      x2gpm_log (LOG_WARNING, "%s error: %s",
                                 conv_func_name, x11_text_conv_strerror(ret));
                      beep ();
                      break;
                    }
                  else if (ret > 0)
                    {
                      x2gpm_log (LOG_INFO, "%s: %d unconvertible chars",
                                 conv_func_name, ret);
                    }
                }

                {
                  int i;
                  for (i = 0; i < n_textlist; i++)
                    if (! tiocsti (textlist[i]))
                      break;
                }

                XFreeStringList (textlist);
                break;

              default:
                break;
              }

            }
        }
    }
}

int
main (int argc, char **argv)
{
  int opt;
  int optindex = 0;
  int option_version = 0;
  int pid;
  FILE *pidfp = NULL;
  int pipefds[2];

  option_display = getenv ("DISPLAY");  /* default */
  saved_argv = argv;

  setbuf (stdout, NULL);
  setbuf (stderr, NULL);
  signal (SIGPIPE, SIG_IGN);

  for (;;)
    {
      static const struct option longopts[] = {
        { "display", required_argument, NULL, 'd' },
        { "pidfile", required_argument, NULL, 'P' },
        { "sync",    no_argument,       NULL, 'S' },
        { "verbose", no_argument,       NULL, 'v' },
        { "debug",   no_argument,       NULL, 'D' },
        { "version", no_argument,       NULL, 'V' },
        { "foreground", no_argument,    NULL, 'F' },
        { NULL }
      };

      opt = getopt_long (argc, argv, "ad:v", longopts, &optindex);
      if (opt == EOF)
        break;

      switch (opt) {
      case 'a':
        option_all = 1;
        break;
      case 'd':
        option_display = optarg;
        break;
      case 'D':
        option_debug = 1;
        break;
      case 'P':
        option_pidfile = optarg;
        break;
      case 'S':
        option_sync = 1;
        break;
      case 'v':
        option_verbose++;
        break;
      case 'V':
        option_version = 1;
        break;
      case 'F':
        option_foreground = 1;
        break;
      case '?':
        exit (1);
      }
    }
  if (option_debug)
    option_verbose = 1;

  if (option_version)
    {
      printf ("x2gpm version %d\n", VERSION);
      if (option_verbose)
        {
          char *sver;
          printf ("gpm library %s\n", Gpm_GetLibVersion (NULL));
          sver = Gpm_GetServerVersion (NULL);
          if (sver)
            printf ("gpm server %s\n", sver);
          else
            printf ("gpm server not installed\n");
        }
      exit (0);
    }

  /* if Xlib doesn't know the locale both XmbTextPropertyToTextList and
     Xutf8TextPropertyToTextList will fail with XLocaleNotSupported, so must
     reset to "C" if the environment variables are no good */
  if (setlocale (LC_ALL, "") == NULL)
    {
      fprintf (stderr, "x2gpm: setlocale error, continuing ASCII-only\n");
    }
  else if (! XSupportsLocale ())
    {
      fprintf (stderr, "x2gpm: locale \"%s\" not supported by Xlib, continuing ASCII-only\n",
               setlocale (LC_ALL, NULL));
      setlocale (LC_ALL, "C");
    }

  if (! option_display)
    {
      fprintf (stderr, "x2gpm: neither --display nor DISPLAY environment variable specified\n");
      exit (1);
    }

  if (option_pidfile != NULL)
    {
      pidfp = fopen (option_pidfile, "w");
      if (pidfp == NULL)
        {
          fprintf (stderr, "x2gpm: error creating pidfile %s: %s\n",
                   option_pidfile, strerror(errno));
          exit (1);
        }
    }

  console_filename = (option_all ? "/dev/console" : "/dev/tty");
  console_fd = open (console_filename, O_WRONLY);
  if (console_fd == -1)
    {
      fprintf (stderr, "x2gpm: cannot open %s: %s",
               console_filename, strerror(errno));
      exit (1);
    }

  if (pipe (pipefds) != 0)
    {
      perror ("x2gpm: cannot create pipe");
      exit (1);
    }

  if (option_verbose && ! option_foreground)
    printf ("forking going into the background\n");

  if (option_foreground)
    {
      pid = 0;
    }
  else
    {
      pid = fork ();
      if (pid == -1)
        {
          perror ("x2gpm: cannot fork");
          exit (1);
        }
    }

  if (pid == 0)
    {
      /* child, or mainline when foreground */

      if (close (pipefds[0]) != 0) /* read side */
        {
          perror ("x2gpm: oops, child error closing read pipe");
          goto CHILD_ERROR;
        }
      if (pidfp != NULL && fclose (pidfp) != 0)
        {
          fprintf (stderr, "oops, error closing pidfile in child\n");
          goto CHILD_ERROR;
        }

      /* catch control-button3 */
      conn.eventMask = GPM_DOWN;
      conn.defaultMask = GPM_DOWN;
      conn.minMod = (1 << KG_CTRL);
      conn.maxMod = (1 << KG_CTRL);
      {
        int fd = Gpm_Open (&conn, option_all ? -1 : 0);
        if (fd < 0)
          {
            fprintf (stderr, "x2gpm: cannot connect to gpm server\n");
          CHILD_ERROR:
            if (! option_foreground)
              {
                int exit_code = 1;
                int ret = write (pipefds[1], &exit_code, sizeof(exit_code));
                if (ret == -1) {
                  perror ("x2gpm: oops, cannot write error code to parent");
                } else if (ret != sizeof(exit_code)) {
                  fprintf (stderr, "x2gpm: oops, cannot write error code to parent (wrote %d bytes)", ret);
                }
              }
            exit (1);
          }
        if (option_verbose)
          printf ("Gpm_Open ok, ret %d gpm_fd %d, vc %d\n",
                  fd, gpm_fd, conn.vc);
      }

      openlog ("x2gpm",
               LOG_CONS,  /* | (option_verbose ? LOG_PERROR : 0), */
               LOG_DAEMON);
      {
        /* suppress LOG_INFO unless verbose
           suppress LOG_DEBUG unless debug */
        int want_bits = ~0;
        if (! option_verbose) want_bits &= ~ LOG_MASK(LOG_INFO);
        if (! option_debug)   want_bits &= ~ LOG_MASK(LOG_DEBUG);
        setlogmask (want_bits);
      }

      if (! option_foreground)
        {
          /* For background operation, dup stdout fd=1 and stderr fd=2 to
             the controlling tty ctermid(), or if no such then the
             console_fd.

             Stdout will be used for the beep() for most alerts to the user.
             There might be stuff in Xlib for or Gpm which writes to stdout
             or stderr so keep something sensible.

             Debian starts daemons with an initial "startpar" with a
             pseudo-tty to collect and print any startup output from a
             background program (and stop parallel startups making a mess of
             each other's outputs).  A background program is no doubt
             supposed to detach itself from its initial stdout/stderr, then
             use whatever syslog or own log file etc.

             Crib note: "lsof -p <PID>", run as root, shows fd of files etc
             held open by process <PID>.
          */
          char *filename = ctermid (NULL);
          int fd = open (filename, O_WRONLY);
          int want_close = 1;
          int out_fd;
          if (fd == -1)
            {
              if (option_verbose)
                {
                  fprintf (stderr, "cannot open controlling terminal %s: %s\n",
                           filename, strerror(errno));
                  fprintf (stderr, "use console instead\n");
                }
              fd = console_fd;
              want_close = 0;
            }
          for (out_fd = 1; out_fd <= 2; out_fd++)
            {
              if (dup2(fd,out_fd) < 0)
                {
                  fprintf (stderr, "cannot dup to stdout/stderr %d: %s\n",
                           out_fd, strerror(errno));
                  goto CHILD_ERROR;
                }
            }
          if (want_close)
            close(fd);
        }

      if (! option_foreground)
        {
          int exit_code = 0;
          int ret = write (pipefds[1], &exit_code, sizeof(exit_code));
          if (ret == -1)
            {
              perror ("x2gpm: oops, cannot write success to parent");
              exit (1);
            }
          if (ret != sizeof(exit_code))
            {
              fprintf (stderr, "x2gpm: oops, cannot write success code to parent (wrote %d bytes)", ret);
              exit (1);
            }
        }

      if (close (pipefds[1]) != 0) /* write side */
        {
          x2gpm_log (LOG_WARNING, "oops, error closing pipe: %s",
                     strerror(errno));
        }

      x2gpm ();
    }
  else
    {
      /* parent */
      int exit_code;

      if (close (pipefds[1]) != 0) /* write side */
        {
          perror ("x2gpm: oops, parent error closing pipe");
          exit (1);
        }

      {
        int ret = read (pipefds[0], &exit_code, sizeof(exit_code));
        if (ret == -1)
          {
            perror ("x2gpm: oops, error reading child exit code");
            exit (1);
          }
        if (ret != sizeof(exit_code))
          {
            fprintf (stderr, "x2gpm: oops, child didn't give exit code (only %d bytes)\n", ret);
            exit (1);
          }
      }
      if (exit_code != 0)
        exit (exit_code);

      x2gpm_log (LOG_INFO, "x2gpm server started, pid %d", pid);

      if (pidfp != NULL)
        {
          if (fprintf (pidfp, "%d", pid) < 0)
            {
              x2gpm_log (LOG_ERR, "error writing pidfile %s: %s",
                         option_pidfile, strerror(errno));
              exit (1);
            }
          if (fclose (pidfp) != 0)
            {
              x2gpm_log (LOG_ERR, "error closing pidfile %s: %s",
                         option_pidfile, strerror(errno));
              exit (1);
            }
        }
    }

  return 0;
}
